@@ -11,8 +11,7 @@ from python_kemptech_api.exceptions import (
     SubVsCannotCreateSubVs,
     RealServerMissingVirtualServiceInfo,
     RealServerMissingLoadmasterInfo,
-    AccessControlMissingLoadmasterInfo,
-    AccessControlMissingVirtualServiceInfo,
+    VirtualServiceACLMissingVirtualServiceInfo,
     ValidationError,
     RuleMissingLoadmasterInfo,
     RangeMissingLoadmasterInfo,
@@ -21,7 +20,9 @@ from python_kemptech_api.exceptions import (
     SiteMissingLoadmasterInfo,
     ClusterMissingLoadmasterInfo,
     CertificateMissingLoadmasterInfo,
-    RangeMaskInvalid, VirtualServiceMissingLoadmasterInfo)
+    CipherListInvalid,
+    RangeMaskInvalid,
+    VirtualServiceMissingLoadmasterInfo)
 from python_kemptech_api.generic import BaseKempObject
 from python_kemptech_api.utils import (
     validate_ip,
@@ -222,9 +223,9 @@ class VirtualService(BaseKempObject):
             return 'Virtual Service {} {}:{} on LoadMaster {}'.format(
                 self.prot.upper(), self.vs, self.port, self.ip_address)
 
-    #@property
-    #def rs(self):
-        #return {int(rs.rsindex): rs for rs in self.get_real_servers()}
+    @property
+    def servers(self):
+        return {int(rs.rsindex): rs for rs in self.get_real_servers()}
 
     def to_api_dict(self):
         api = super(VirtualService, self).to_api_dict()
@@ -363,26 +364,39 @@ class VirtualService(BaseKempObject):
         return RealServer(self.access_info, ip, port)
 
     def create_access_control(self, addvs, addr):
-        """AccessControl factory with pre-configured LoadMaster connection."""
-        # AccessControl requires the full vs ID parameters
-        # in order to operate correctly, so if an index is
-        # currently populated in this instance of VirtualService
-        # it will instantiate AccessControl with non valid params.
-        # Temporarily making index None will bypass this behaviour
-        # and pass the required parameters along to AccessControl.
-        id = self.index
-        self.index = None
-        try:
-            acl = AccessControl(self.access_info, addvs, addr)
-        except AccessControlMissingLoadmasterInfo:
-            raise
-        finally:
-            self.index = id
+        """AccessControl factory with pre-configured LoadMaster connection and
+        ACL definition."""
+
+        log.info("This method has been deprecated, please manipualte ACL "
+                 "objects directly")
+
+        acl = self.acl
+
+        if addvs == "black":
+            acl.blacklist[addr] = ""
+        elif addvs == "white":
+            acl.whitelist[addr] = ""
+        else:
+            log.warning("ACL List %s is not valid, ACLs have not been modified",
+                        addvs)
+
+        acl.update()
 
         return acl
 
+    @property
+    def acl(self):
+        return self.get_vs_acl()
+
+    def get_vs_acl(self):
+        access_info = self.access_info
+        access_info["ip"] = self.vs
+        access_info["port"] = self.port
+        access_info["prot"] = self.prot
+
+        return VirtualServiceACL(self.access_info)
+
     def get_real_server(self, real_server_address=None, real_server_port=None):
-        validate_ip(real_server_address)
         validate_port(real_server_port)
 
         if self.index is None:
@@ -484,6 +498,18 @@ class VirtualService(BaseKempObject):
         self.cachepercent = service.get('CachePercent', None)
 
         self.sslacceleration = service.get('SSLAcceleration', None)
+        self.sslrewrite = service.get('SSLRewrite', None)
+        self.sslreverse = service.get('SSLReverse', None)
+        self.sslreencrypt = service.get('SSLReencrypt', None)
+        self.starttlsmode = service.get('StartTLSMode', None)
+        self.tlstype = service.get('TlsType', None)
+        self.cipherset = service.get('CipherSet', None)
+        self.certfile = service.get('CertFile', None)
+        self.clientcert = service.get('ClientCert', None)
+        self.ocspverify = service.get('OCSPVerify', None)
+        self.reversesnihostname = service.get('ReverseSNIHostname', None)
+        self.needhostname = service.get('NeedHostName', None)
+
         if self.sslacceleration is None or self.sslacceleration == 'N':
             self._ssl = False
         else:
@@ -496,6 +522,7 @@ class VirtualService(BaseKempObject):
             self.cipherset = None
             self.ciphers = None
             self.needhostname = None
+
         else:
             # Rewrite the SSL Rewrite value based on the table:
             # SSL Rewrite cannot be set as an integer, even
@@ -515,6 +542,7 @@ class VirtualService(BaseKempObject):
 
             #
             if self.certfile is not None:
+                log.info("Splitting certfile field into a list")
                 self.certfile = str(self.certfile).split()
             else:
                 self.certfile = []
@@ -595,7 +623,6 @@ class RealServer(BaseKempObject):
         "dnsname": "DnsName"
     }
 
-    """
     @property
     def rs(self):
         if hasattr(self, "dnsname") and self.dnsname is not None:
@@ -611,12 +638,12 @@ class RealServer(BaseKempObject):
             self.dnsname = value
         else:
             self.addr = value
-    """
 
     def to_api_dict(self):
         # Populate RS field into dictionary manually
         as_dict = super(RealServer, self).to_api_dict()
         as_dict['rs'] = self.rs
+        as_dict.pop('addr')
         return as_dict
 
     def __init__(self, loadmaster_virt_service_info, ip, port=80):
@@ -649,66 +676,211 @@ class RealServer(BaseKempObject):
     def __str__(self):
         return 'Real Server {} on {}'.format(self.rs, self.vs)
 
+    def save(self, update=False):
+        super(RealServer, self).save(update)
+        self.refresh()
+
+
+class BaseACLObject(BaseKempObject):
+    _API = "/aclcontrol"
+    _API_ADD = ""
+    _API_DEL = ""
+    _API_LIST = ""
 
-class AccessControl(BaseKempObject):
-    _API_ADD = "/aclcontrol"
     API_INIT_PARAMS = {
-        "ip": "VSAddress",
-        "port": "VSPort",
-        "prot": "Protocol"
     }
     _API_BASE_PARAMS = [
-        "vsip",
-        "vsport",
-        "vsprot",
-        "addr",
-        "addvs"
     ]
     _API_DEFAULT_ATTRIBUTES = {
-        "vsip",
-        "vsport",
-        "vsprot"
     }
 
-    def __init__(self, loadmaster_virt_service_info, addvs, addr):
+    def __init__(self, loadmaster_info):
+        # Attach to the LoadMaster
 
-        try:
-            self.vsip = loadmaster_virt_service_info["vs"]
-        except KeyError:
-            raise AccessControlMissingVirtualServiceInfo("vs")
+        self.appliance = loadmaster_info['appliance']
+        self.blacklist = {}
+        self.whitelist = {}
 
-        self.vsport = loadmaster_virt_service_info.get("port", None)
-        self.vsprot = loadmaster_virt_service_info.get("prot", None)
+        super(BaseACLObject, self).__init__(loadmaster_info)
 
-        self.addvs = addvs
-        self.addr = addr
-
-        try:
-            self.endpoint = loadmaster_virt_service_info["endpoint"]
-        except KeyError:
-            raise AccessControlMissingLoadmasterInfo("endpoint")
-
-        try:
-            self.ip_address = loadmaster_virt_service_info["ip_address"]
-        except KeyError:
-            raise AccessControlMissingLoadmasterInfo("ip_address")
-
-        super(AccessControl, self).__init__(loadmaster_virt_service_info)
-
-    def __str__(self):
-        return 'Access Control on {}'.format(self.vsip)
+        self.refresh()
 
     def save(self, update=False):
-        if not update:
-            response = self._get(self._API_ADD,
-                                 self._get_base_parameters())
-        else:
-            response = self._get(self._API_LIST,
-                                 self._get_default_attributes())
+        self._sync()
+        self.refresh()
+
+    def _sync(self):
+        # Sync the blacklist and whitelist to the LoadMaster
+
+        # Grab the new data and save it before refreshing to get the old data
+        new_blacklist = self.blacklist
+        new_whitelist = self.whitelist
+        self.refresh()
+        old_blacklist = self.blacklist
+        old_whitelist = self.whitelist
+
+        # Handle the blacklist changes
+        for address, comment in old_blacklist.items():
+            if address not in new_blacklist.keys():
+                self._delete_entry('black', address)
+            else:
+                if new_blacklist[address] != comment:
+                    self._delete_entry('black', address)
+                    self._add_entry('black', address, new_blacklist[address])
+
+        for address, comment in {key: value for key, value in new_blacklist.items()
+                                 if key not in old_blacklist.keys()}.items():
+            self._add_entry('black', address, comment)
+
+        # Now handle the whitelist
+        for address, comment in old_whitelist.items():
+            if address not in new_whitelist.keys():
+                self._delete_entry('white', address)
+            else:
+                if new_whitelist[address] != comment:
+                    self._delete_entry('white', address)
+                    self._add_entry('white', address, new_whitelist[address])
+
+        for address, comment in {key: value for key, value in new_whitelist.items()
+                                 if key not in old_whitelist.keys()}.items():
+            self._add_entry('white', address, comment)
+
+    def _add_entry(self, list_type, address, comment=None):
+        parameters = self._get_base_parameters()
+
+        parameters[self._API_ADD] = list_type
+        parameters['addr'] = address
+
+        # Only valid on 7.2.37.0 and higher
+        if self.appliance['version'] >= "7.2.37.0":
+            parameters['comment'] = comment
+
+        response = self._get(  # pylint: disable=protected-access
+            self._API,
+            parameters)
 
         if not is_successful(response):
             raise KempTechApiException(get_error_msg(response))
 
+    def _delete_entry(self, list_type, address):
+        parameters = self._get_base_parameters()
+
+        parameters[self._API_DEL] = list_type
+        parameters['addr'] = address
+
+        response = self._get(  # pylint: disable=protected-access
+            self._API,
+            parameters)
+
+        if not is_successful(response):
+            raise KempTechApiException(get_error_msg(response))
+
+    def refresh(self):
+        # First handle whitelist
+        parameters = self._get_base_parameters()
+
+        parameters[self._API_LIST] = "white"
+
+        whitelist_response = self._get(  # pylint: disable=protected-access
+            self._API,
+            parameters)
+
+        whitelist_data = get_data(whitelist_response)
+
+        if isinstance(self, VirtualServiceACL):
+            whitelist_xml = whitelist_data['VS']['Whitelist']
+        else:
+            whitelist_xml = whitelist_data['Whitelist']
+
+        if whitelist_xml is None:
+            self.whitelist = {}
+        # Handle pre-7.2.37.0 cases
+        elif "addr" in whitelist_xml.keys():
+            self.whitelist = {
+                address: "" for
+                address in cast_to_list(whitelist_xml['addr'])
+            }
+        else:
+            self.whitelist = {
+                ip['addr']: ip['comment'] or ""
+                for ip in cast_to_list(whitelist_xml['IP'])
+            }
+
+        # Next verse, same as the first!
+        parameters = self._get_base_parameters()
+
+        parameters[self._API_LIST] = "black"
+
+        blacklist_response = self._get(  # pylint: disable=protected-access
+            self._API,
+            parameters)
+
+        blacklist_data = get_data(blacklist_response)
+
+        if isinstance(self, VirtualServiceACL):
+            blacklist_xml = blacklist_data['VS']['Blacklist']
+        else:
+            blacklist_xml = blacklist_data['Blacklist']
+
+        if blacklist_xml is None:
+            self.blacklist = {}
+        # Handle pre-7.2.37.0 cases
+        elif "addr" in blacklist_xml.keys():
+            self.blacklist = {
+                address: ""
+                for address in cast_to_list(blacklist_xml['addr'])
+            }
+        else:
+            self.blacklist = {
+                ip['addr']: ip['comment'] or ""
+                for ip in cast_to_list(blacklist_xml['IP'])
+            }
+
+
+class GlobalACL(BaseACLObject):
+    _API_ADD = "add"
+    _API_DEL = "del"
+    _API_LIST = "list"
+
+    def __repr__(self):
+        return 'Global ACL on {}'.format(self.appliance)
+
+
+class VirtualServiceACL(BaseACLObject):
+    _API_ADD = "addvs"
+    _API_DEL = "delvs"
+    _API_LIST = "listvs"
+
+    _API_BASE_PARAMS = [
+        "vsip",
+        "vsport",
+        "vsprot"
+    ]
+
+    def __init__(self, loadmaster_virt_service_info):
+        try:
+            self.vsip = loadmaster_virt_service_info["vs"]
+        except KeyError:
+            raise VirtualServiceACLMissingVirtualServiceInfo("vs")
+
+        try:
+            self.vsport = loadmaster_virt_service_info.get("port", None)
+        except KeyError:
+            raise VirtualServiceACLMissingVirtualServiceInfo("port")
+
+        try:
+            self.vsprot = loadmaster_virt_service_info.get("prot", None)
+        except KeyError:
+            raise VirtualServiceACLMissingVirtualServiceInfo("prot")
+
+        super(VirtualServiceACL, self).__init__(loadmaster_virt_service_info)
+
+    def __repr__(self):
+        return 'Virtual Service ACL on {}/{}:{}'.format(
+            self.vsprot,
+            self.vsip,
+            self.vsport)
+
 
 class Template(BaseKempObject):
     _API_ADD = ""
@@ -955,6 +1127,7 @@ class Fqdn(BaseKempObject):
             pass
 
         super(Fqdn, self).save(update)
+        self.refresh()
 
     def populate_default_attributes(self, dictionary):
         super(Fqdn, self).populate_default_attributes(dictionary)
@@ -967,7 +1140,7 @@ class Fqdn(BaseKempObject):
             if self.failtime > 0 and self.failtime % 60 == 0:
                 # Convert from seconds to minutes
                 self.failtime = int(self.failtime / 60)
-        except TypeError:
+        except (TypeError, AttributeError):
             self.failtime = None
 
     @property
@@ -1010,7 +1183,7 @@ class Fqdn(BaseKempObject):
         try:
             response = self._get(self._API_LIST, fqdn)
             data = get_data(response)
-            xml_object = data["fqdn"].get(self.API_TAG, [])
+            xml_object = data[self.API_TAG].get(Site.API_TAG, [])
         except KempTechApiException:
             xml_object = []
 
@@ -1020,8 +1193,9 @@ class Fqdn(BaseKempObject):
         # ValidationError, which is the best we can do for now
         # (without changing the upstream code and raising an
         # exception earlier, possibly retrying)
-        if not isinstance(xml_object, list):
-            xml_object = [xml_object]
+
+        xml_object = cast_to_list(xml_object)
+
         for x in xml_object:
             obj = self.build_site(x)
             obj_list.append(obj)
@@ -1031,7 +1205,7 @@ class Fqdn(BaseKempObject):
         """Create a object instance with standard defaults"""
         build_parameters = {}
 
-        for parameter, tag in self.API_INIT_PARAMS.items():
+        for parameter, tag in Site.API_INIT_PARAMS.items():
             build_parameters[parameter] = site.get(tag)
 
         obj = Site(self.access_info, **build_parameters)
@@ -1385,11 +1559,15 @@ class Cluster(BaseKempObject):
             return 'Cluster #{} {} at {} on LoadMaster {}'.format(
                 self.id, self.name, self.ip, self.ip_address)
 
+    def save(self, update=False):
+        super(Cluster, self).save(update)
+        self.refresh()
+
     def populate_default_attributes(self, dictionary):
         super(Cluster, self).populate_default_attributes(dictionary)
 
         # Clear checkerport if it's not in use
-        if self.checkerport == "0":
+        if hasattr(self, "checkerport") and self.checkerport == "0":
             if self.checker != "tcp":
                 self.checkerport = None
             else:
@@ -1494,6 +1672,9 @@ class Range(BaseKempObject):
             if not is_successful(response):
                 raise KempTechApiException(get_error_msg(response))
 
+            # We need to refresh here, creating the range does not return data
+            self.refresh()
+
         # Set Coordinates
         if self.lat is not None and self.long is not None:
             response = self._get(self._API_MOD_LOC,
@@ -1596,16 +1777,65 @@ class CustomLocation(BaseKempObject):
         return send_response(response)
 
 
+class CipherSet(BaseKempObject):
+    def __init__(self, loadmaster_info, cipherset_name, ciphers):
+        self.cipherset_name = cipherset_name
+        cipher_regex = re.compile("^([A-Z0-9-]*:*)*[^:]$")
+        if isinstance(ciphers, list):
+            self.ciphers = ":".join(ciphers)
+        elif isinstance(ciphers, str) and cipher_regex.match(ciphers):
+            self.ciphers = ciphers
+        else:
+            raise CipherListInvalid(ciphers)
+
+        try:
+            self.endpoint = loadmaster_info["endpoint"]
+        except KeyError:
+            raise CertificateMissingLoadmasterInfo("endpoint")
+
+        try:
+            self.ip_address = loadmaster_info["ip_address"]
+        except KeyError:
+            raise CertificateMissingLoadmasterInfo("ip_address")
+
+        super(CipherSet, self).__init__(loadmaster_info)
+
+    def __str__(self):
+        return 'Cipher List {}'.format(self.ciphers)
+
+    def _get_base_parameters(self):
+        """Returns the bare minimum cipherset parameters"""
+        return {
+            "name": self.cipherset_name,
+            "value": self.ciphers
+        }
+
+    def save(self, update=False):
+        response = self._get('/modifycipherset',
+                             parameters=self._get_base_parameters())
+        if is_successful(response):
+            pass
+        else:
+            raise KempTechApiException(get_error_msg(response))
+
+    def delete(self):
+        response = self._get("/delcipherset", self._get_base_parameters())
+        return send_response(response)
+
+
 class Certificate(BaseKempObject):
-    def __init__(self, loadmaster_info, certname, certfile=None, certpass=None,
-                 type=None):
+    def __init__(self, loadmaster_info, certname,
+                 certfile=None, certpass=None):
         self.certname = certname
 
+        # If certname is a structure, pull out the name and set the modulus
+        if isinstance(self.certname, dict):
+            self.modulus = self.certname['modulus']
+            self.certname = self.certname['name']
+
         if certfile is not None:
             self.certfile = certfile
 
-        self.type = type or "cert"
-
         if certpass is not None:
             self.certpass = certpass
         else:
@@ -1642,26 +1872,73 @@ class Certificate(BaseKempObject):
             }
 
     def save(self, update=False):
-        if self.type == 'cert':
-            response = self._post("/addcert", file=self.certfile,
-                                  parameters=self._get_base_parameters())
-        elif self.type == 'intermediate':
-            response = self._post("/addintermediate", file=self.certfile,
-                                  parameters=self._get_base_parameters())
-        else:
-            pass
+        response = self._post("/addcert", file=self.certfile,
+                              parameters=self._get_base_parameters())
+
         if is_successful(response):
             pass
         else:
             raise KempTechApiException(get_error_msg(response))
 
     def delete(self):
-        if self.type == 'cert':
-            response = self._get("/delcert",
-                                 self._get_base_parameters())
-        elif self.type == 'intermediate':
-            response = self._get("/delintermediate",
-                                 self._get_base_parameters())
+        response = self._get("/delcert",
+                             self._get_base_parameters())
+
+        return send_response(response)
+
+    def populate_default_attributes(self, dictionary):
+        super(Certificate, self).populate_default_attributes(dictionary)
+
+        # If certname is a structure, pull out the name and set the modulus
+        if isinstance(self.certname, dict):
+            self.modulus = self.certname['modulus']
+            self.certname = self.certname['name']
+
+
+class IntermediateCertificate(BaseKempObject):
+    def __init__(self, loadmaster_info, certname, certfile=None):
+        self.certname = certname
+
+        # If certname is a structure, pull out the name and set the modulus
+        if isinstance(self.certname, dict):
+            self.modulus = self.certname['modulus']
+            self.certname = self.certname['name']
+
+        if certfile is not None:
+            self.certfile = certfile
+
+        try:
+            self.endpoint = loadmaster_info["endpoint"]
+        except KeyError:
+            raise CertificateMissingLoadmasterInfo("endpoint")
+        try:
+            self.ip_address = loadmaster_info["ip_address"]
+        except KeyError:
+            raise CertificateMissingLoadmasterInfo("ip_address")
+
+        super(IntermediateCertificate, self).__init__(loadmaster_info)
+
+    def __str__(self):
+        return 'Intermediate Certificate {}'.format(self.certname)
+
+    def _get_base_parameters(self):
+        """Returns the bare minimum VS parameters. IP, port and protocol"""
+        return {
+            "cert": self.certname,
+        }
+
+    def save(self, update=False):
+        response = self._post("/addintermediate",
+                              file=self.certfile,
+                              parameters=self._get_base_parameters())
+
+        if is_successful(response):
+            pass
+        else:
+            raise KempTechApiException(get_error_msg(response))
+
+    def delete(self):
+        response = self._get("/delintermediate", self._get_base_parameters())
         return send_response(response)
 
 
@@ -1781,22 +2058,29 @@ class Interface(BaseKempObject):
         super(Interface, self).populate_default_attributes(params)
 
         # Strip additional addresses into a list
-        if isinstance(self.additionaladdresses, OrderedDict):
-            self.additionaladdresses = self.additionaladdresses['IPaddress']
-
-            if not isinstance(self.additionaladdresses, list):
-                self.additionaladdresses = [self.additionaladdresses]
+        if not hasattr(self, "additionaladdresses"):
+            self.additionaladdresses = []
         elif self.additionaladdresses is None:
             self.additionaladdresses = []
+        elif isinstance(self.additionaladdresses, OrderedDict):
+            self.additionaladdresses = self.additionaladdresses['IPaddress']
+            self.additionaladdresses = cast_to_list(self.additionaladdresses)
 
-        if self.geotraffic == "no":
+        if not hasattr(self, "geotraffic"):
+            self.geotraffic = None
+        elif self.geotraffic == "no":
             self.geotraffic = "0"
         elif self.geotraffic == "yes":
             self.geotraffic = "1"
 
+        self._additionaladdresses = []
+
+        for address in self.additionaladdresses:
+            self._additionaladdresses.append(address)
+
     def set_additionaladdresses(self):
         new = self.additionaladdresses
-        old = Interface.get(self, self.interface).additionaladdresses
+        old = self._additionaladdresses
 
         for address in list(set(old) - set(new)):
             self._delete_additionaladdress(address)
